<?php
// Copyright (C) <2015-present>  <it-novum GmbH>
//
// This file is licensed under the terms of the openITCOCKPIT Enterprise Edition license agreement.
// The license agreement and license key were sent with the order confirmation.

declare(strict_types=1);

namespace App\Command;

use App\Model\Entity\PushAgent;
use App\Model\Table\PushAgentsTable;
use Cake\Command\Command;
use Cake\Console\Arguments;
use Cake\Console\ConsoleIo;
use Cake\Console\ConsoleOptionParser;
use Cake\I18n\FrozenTime;
use Cake\ORM\TableRegistry;

/**
 * Agent command.
 * This command is used by the NSTA to handle Push Agent configuration on the Satellite System
 * This command is only used for Agents running in Push Mode !
 */
class AgentCommand extends Command {

    const RC_UP = 0;
    const RC_DOWN = 1;
    const RC_UNREACHABLE = 2;

    const RC_OK = 0;
    const RC_WARNING = 1;
    const RC_CRITICAL = 2;
    const RC_UNKNOWN = 3;

    /**
     * Hook method for defining this command's option parser.
     *
     * @see https://book.cakephp.org/4/en/console-commands/commands.html#defining-arguments-and-options
     * @param \Cake\Console\ConsoleOptionParser $parser The parser to be defined
     * @return \Cake\Console\ConsoleOptionParser The built parser.
     */
    public function buildOptionParser(ConsoleOptionParser $parser): ConsoleOptionParser {
        $parser = parent::buildOptionParser($parser);

        // Agent CLI API Options (Satellite only)
        $parser->addOption('get-push-agents', [
            'help'    => __('Get a list of available push agents for select_agent step'),
            'boolean' => true,
            'default' => false
        ]);

        $parser->addOption('get-all-push-agents', [
            'help'    => __('Get a list of all push agents'),
            'boolean' => true,
            'default' => false
        ]);

        $parser->addOption('assign', [
            'help'    => __('Assign an push-agent-id to an host-uuid'),
            'boolean' => true,
            'default' => false
        ]);

        $parser->addOption('push-agent-id', [
            'help'    => __('The ID of the push agent you want to assign'),
            'default' => 0
        ]);

        $parser->addOption('host-uuid', [
            'help'    => __('The host UUID you want to assign to an push agent'),
            'default' => ''
        ]);

        $parser->addOption('query-agent', [
            'help'    => __('Query the output of an Agent running in Push Mode'),
            'boolean' => true,
            'default' => false
        ]);

        $parser->addOption('assign-from-json', [
            'help'    => __('Assign multiple host-uuids to push-agent-ids'),
            'boolean' => true,
            'default' => false
        ]);

        $parser->addOption('path', [
            'help'    => __('Path to json file with host-uuids to push-agent-ids assignments'),
            'default' => '/opt/openitc/nagios/etc/push_agents_import_module.json'
        ]);

        // Check options (Same as on Master System)
        // Freshness check options
        $parser->addOption('check', [
            'help'    => 'Determines the host state of an host running in Push Mode by evaluating the timestamp of the last received check  results.',
            'boolean' => true
        ]);
        $parser->addOption('host-check', [
            'short'   => 'H',
            'help'    => 'Determines if this check runs as host or service check. If set, only the critical value will be evaluated.',
            'boolean' => true
        ]);
        $parser->addOption('hostuuid', [
            'help' => 'The UUID of the host to evaluate. Same as --host-uuid'
        ]);
        $parser->addOption('warning', [
            'short'   => 'w',
            'help'    => 'Warning threshold in seconds.',
            'default' => 60
        ]);
        $parser->addOption('critical', [
            'short'   => 'c',
            'help'    => 'Critical threshold in seconds.',
            'default' => 120
        ]);

        return $parser;
    }

    /**
     * Implement this method with your command's logic.
     *
     * @param \Cake\Console\Arguments $args The command arguments.
     * @param \Cake\Console\ConsoleIo $io The console io
     * @return null|void|int The exit code or null for success
     */
    public function execute(Arguments $args, ConsoleIo $io) {
        // To be compatible with the Master System we have --host-uuid and --hostuuid
        $hostUuid = $args->getOption('host-uuid');
        if(empty($hostUuid)){
            $hostUuid = $args->getOption('hostuuid');
        }

        if ($args->getOption('get-push-agents') === true) {
            $this->printPushAgents($io, $hostUuid);
        }

        if ($args->getOption('get-all-push-agents') === true) {
            $this->printAllPushAgents($io);
        }

        if ($args->getOption('assign') === true) {
            $this->assignPushAgentToHost(
                $io,
                (int)$args->getOption('push-agent-id'),
                $hostUuid
            );
        }

        if ($args->getOption('query-agent') === true) {
            $this->getAgentOutput(
                $io,
                $hostUuid
            );
        }

        if ($args->getOption('assign-from-json') === true) {
            $this->assignFromJson(
                $io,
                $args->getOption('path')
            );
        }

        // Running as Check Plugin
        if ($args->getOption('check')) {
            $this->checkPushAgentFreshness($args, $io, $hostUuid);
        }
    }

    public function printPushAgents(ConsoleIo $io, ?string $hostUuid) {
        /** @var PushAgentsTable $PushAgentsTable */
        $PushAgentsTable = TableRegistry::getTableLocator()->get('PushAgents');
        echo json_encode($PushAgentsTable->getPushAgentsForAssignments($hostUuid));
    }

    public function assignPushAgentToHost(ConsoleIo $io, int $pushAgentId, string $hostUuid) {
        if ($pushAgentId <= 0) {
            $io->error('--push-agent-id is missing or wrong');
            exit(1);
        }

        if (strlen($hostUuid) !== 36) {
            $io->error('--host-uuid is missing or wrong');
            exit(1);
        }

        /** @var PushAgentsTable $PushAgentsTable */
        $PushAgentsTable = TableRegistry::getTableLocator()->get('PushAgents');

        $pushAgent = $PushAgentsTable->get($pushAgentId);
        $pushAgent->set('host_uuid', strtolower($hostUuid));
        $PushAgentsTable->save($pushAgent);
    }

    public function getAgentOutput(ConsoleIo $io, string $hostUuid) {
        if (strlen($hostUuid) !== 36) {
            $io->error('--host-uuid is missing or wrong');
            exit(1);
        }

        /** @var PushAgentsTable $PushAgentsTable */
        $PushAgentsTable = TableRegistry::getTableLocator()->get('PushAgents');

        $result = $PushAgentsTable->find()
            ->select([
                'id',
                'checkresults'
            ])
            ->where(['host_uuid' => $hostUuid])
            ->disableHydration()
            ->firstOrFail();

        echo $result['checkresults'];
    }

    /**
     * This is used by the Import Module
     *
     * @param ConsoleIo $io
     * @return void
     */
    public function printAllPushAgents(ConsoleIo $io) {
        /** @var PushAgentsTable $PushAgentsTable */
        $PushAgentsTable = TableRegistry::getTableLocator()->get('PushAgents');
        echo json_encode($PushAgentsTable->getAllPushAgents());
    }


    public function assignFromJson(ConsoleIo $io, string $path) {
        if (empty($path)) {
            $io->error('--path is missing or wrong');
            exit(1);
        }

        if (!file_exists($path)) {
            $io->error('no such file or --path is wrong');
            exit(1);
        }

        $io->verbose('Start assignment from push_agents_import_module.json file');

        $json = file_get_contents($path);
        $json = json_decode($json, true);

        /** @var PushAgentsTable $PushAgentsTable */
        $PushAgentsTable = TableRegistry::getTableLocator()->get('PushAgents');

        foreach ($json as $assignments) {
            $pushAgent = $PushAgentsTable->getPushAgentsByUuid($assignments['uuid']);
            if ($pushAgent) {
                $pushAgent->set('host_uuid', $assignments['_matchingData']['Hosts']['uuid']);
                $PushAgentsTable->save($pushAgent);
            }
        }
    }

    /**
     * @param Arguments $args
     * @param ConsoleIo $io
     */
    private function checkPushAgentFreshness(Arguments $args, ConsoleIo $io, ?string $uuid) {
        $warning = (int)$args->getOption('warning');
        $critical = (int)$args->getOption('critical');
        $isHostcheck = $args->getOption('host-check') === true;

        if (empty($uuid)) {
            $rc = self::RC_UNREACHABLE;
            if (!$isHostcheck) {
                $rc = self::RC_UNKNOWN;
            }

            $this->nagOutput('No host uuid given', $rc);
        }

        /** @var PushAgentsTable $PushAgentsTable */
        $PushAgentsTable = TableRegistry::getTableLocator()->get('PushAgents');

        /** @var PushAgent $record */
        $record = $PushAgentsTable->getPushAgentForFreshnessCheck($uuid);
        if (empty($record)) {
            $rc = self::RC_UNREACHABLE;
            if (!$isHostcheck) {
                $rc = self::RC_UNKNOWN;
            }

            $this->nagOutput('No push agent configured for given host.', $rc);
        }

        $lastUpdate = null;
        if ($record->last_update instanceof FrozenTime) {
            $lastUpdate = $record->last_update->getTimestamp();
        }

        if ($lastUpdate === null) {
            if (empty($record)) {
                if ($isHostcheck) {
                    $this->nagOutput('No last update timestamp found in database', self::RC_UNREACHABLE);
                }
                $this->nagOutput('No last update timestamp found in database', self::RC_UNKNOWN);
            }
        }

        $now = time();
        $diff = $now - $lastUpdate;

        $rc = self::RC_UP;
        $output = sprintf('Up: Last check update received from Agent at: %s | age=%ss',
            date('c', $lastUpdate),
            $diff
        );
        if ($isHostcheck === true) {
            if ($diff >= $critical) {
                $output = sprintf('Down: Last received check update from Agent is older than %s seconds: %s | age=%ss',
                    $diff,
                    date('c', $lastUpdate),
                    $diff
                );
                $rc = self::RC_DOWN;
            }

            $this->nagOutput($output, $rc);
        }

        // Running as service check
        $rc = self::RC_OK;
        $output = sprintf('Ok: Last check update received from Agent at: %s | age=%ss;%s;%s',
            date('c', $lastUpdate),
            $diff,
            $warning,
            $critical
        );
        if ($diff >= $critical) {
            $output = sprintf('Critical: Last received check update from Agent is older than %s seconds: %s | age=%ss;%s;%s',
                $diff,
                date('c', $lastUpdate),
                $diff,
                $warning,
                $critical
            );
            $rc = self::RC_CRITICAL;
            $this->nagOutput($output, $rc);
        }

        $this->nagOutput($output, $rc);

        if ($diff >= $warning) {
            $output = sprintf('Warning: Last received check update from Agent is older than %s seconds: %s | age=%ss;%s;%s',
                $diff,
                date('c', $lastUpdate),
                $diff,
                $warning,
                $critical
            );
            $rc = self::RC_WARNING;
            $this->nagOutput($output, $rc);
        }

        // Ok
        $this->nagOutput($output, $rc);
    }

    /**
     * @param string $output
     * @param int $rc
     */
    private function nagOutput($output, $rc) {
        echo $output;
        echo "\n";
        exit($rc);

    }
}
