<?php
// Copyright (C) <2015-present>  <it-novum GmbH>
//
// This file is licensed under the terms of the openITCOCKPIT Enterprise Edition license agreement.
// The license agreement and license key were sent with the order confirmation.
declare(strict_types=1);

namespace App\Command;

use App\Lib\SystemConfig;
use Cake\Console\Arguments;
use Cake\Command\Command;
use Cake\Console\ConsoleIo;
use Cake\Console\ConsoleOptionParser;
use Twig\Environment;
use Twig\Loader\FilesystemLoader;

/**
 * ConfigGenerator command.
 */
class ConfigGeneratorCommand extends Command {

    /**
     * @var ConsoleIo
     */
    private $io;

    /**
     * @var string
     */
    private $configBase;

    /**
     * @var SystemConfig
     */
    private $SystemConfig;

    /**
     * Hook method for defining this command's option parser.
     *
     * @see https://book.cakephp.org/3.0/en/console-and-shells/commands.html#defining-arguments-and-options
     *
     * @param \Cake\Console\ConsoleOptionParser $parser The parser to be defined
     * @return \Cake\Console\ConsoleOptionParser The built parser.
     */
    public function buildOptionParser(ConsoleOptionParser $parser): ConsoleOptionParser {
        $parser = parent::buildOptionParser($parser);

        $parser->addOption('statusengine-worker', [
            'help'    => __('Generate configuration file for Statusengine 3 Worker'),
            'boolean' => true,
            'default' => false
        ]);

        $parser->addOption('statusengine-broker', [
            'help'    => __('Generate configuration file for Statusengine 3 Broker Module'),
            'boolean' => true,
            'default' => false
        ]);

        $parser->addOption('carbon-stack', [
            'help'    => __('Generate configuration files for the Carbon Stack'),
            'boolean' => true,
            'default' => false
        ]);

        $parser->addOption('get-perfdata-backend', [
            'help'    => __('Returns the currently used perfdata backend as string. whisper or mysql'),
            'boolean' => true,
            'default' => false
        ]);

        return $parser;
    }

    /**
     * Implement this method with your command's logic.
     *
     * @param \Cake\Console\Arguments $args The command arguments.
     * @param \Cake\Console\ConsoleIo $io The console io
     * @return null|void|int The exit code or null for success
     */
    public function execute(Arguments $args, ConsoleIo $io) {
        $this->io = $io;

        $this->configBase = ROOT . DS . 'system' . DS;
        $this->SystemConfig = new SystemConfig();

        if ($args->getOption('statusengine-worker') === true) {
            $this->generateStatusengineWorkerConfig();
        }

        if ($args->getOption('statusengine-broker') === true) {
            $this->generateStatusengineBrokerModuleConfig();
        }

        if ($args->getOption('carbon-stack') === true) {
            $this->generateCarbonConfigs();
        }

        if ($args->getOption('get-perfdata-backend') === true) {
            if ($this->SystemConfig->getPerfdataBackend()->isWhisper()) {
                $this->io->out('whisper');
                $this->abort(0);
            }

            if ($this->SystemConfig->getPerfdataBackend()->isMySQL()) {
                $this->io->out('mysql');
                $this->abort(0);
            }
        }
    }

    private function generateStatusengineWorkerConfig() {
        $template = $this->configBase . 'Statusengine3' . DS . 'config.yml';
        $outfile = '/opt/openitc/statusengine3/worker/etc/config.yml';
        $this->io->out(sprintf('Generate %s', $outfile));

        $ini_file = parse_ini_file('/opt/openitc/etc/mysql/mysql.cnf');

        $perfdata_backend = 'mysql';
        if ($this->SystemConfig->getPerfdataBackend()->isWhisper()) {
            $perfdata_backend = 'graphite';
        }

        $config = [
            'mysql_server'   => $ini_file['host'],
            'mysql_port'     => $ini_file['port'],
            'mysql_username' => $ini_file['user'],
            'mysql_password' => $ini_file['password'],
            'mysql_database' => $ini_file['database'],

            'perfdata_backend' => $perfdata_backend,

            'graphite_prefix'  => $this->SystemConfig->getGraphitePrefix(),
            'graphite_address' => $this->SystemConfig->getGraphiteAddress(),
            'graphite_port'    => $this->SystemConfig->getGraphitePort(),

            'age_hostchecks'               => $this->SystemConfig->getAgeHostchecks(),
            'age_host_acknowledgements'    => $this->SystemConfig->getAgeHostAcknowledgements(),
            'age_host_notifications'       => $this->SystemConfig->getAgeHostNotifications(),
            'age_host_statehistory'        => $this->SystemConfig->getAgeHostStatehistory(),
            'age_host_downtimes'           => $this->SystemConfig->getAgeHostDowntimes(),
            'age_servicechecks'            => $this->SystemConfig->getAgeServicechecks(),
            'age_service_acknowledgements' => $this->SystemConfig->getAgeServiceAcknowledgements(),
            'age_service_notifications'    => $this->SystemConfig->getAgeServiceNotifications(),
            'age_service_statehistory'     => $this->SystemConfig->getAgeServiceStatehistory(),
            'age_service_downtimes'        => $this->SystemConfig->getAgeServiceDowntimes(),
            'age_logentries'               => $this->SystemConfig->getAgeLogentries(),
            'age_perfdata'                 => $this->SystemConfig->getAgePerfdata()
        ];

        $loader = new FilesystemLoader(dirname($template));
        $twig = new Environment($loader);

        file_put_contents($outfile, $twig->render(basename($template), $config));
    }

    private function generateCarbonConfigs() {
        $config = [
            // storage-schemas.conf
            'graphite_storage_schema' => $this->SystemConfig->getGraphiteStorageSchema(),

            // local_settings.py
            'graphite_timezone'       => $this->SystemConfig->getGraphiteTimezone(),

            // docker-compose
            'docker_auto_networking'  => $this->SystemConfig->getDockerAutoNetworkingEnabled(),
            'docker_compose_subnet'   => $this->SystemConfig->getDockerComposeSubnet()
        ];

        $configFiles = [
            $this->configBase . 'opt/openitc/etc/carbon' . DS . 'storage-schemas.conf' => '/opt/openitc/etc/carbon/storage-schemas.conf',
            $this->configBase . 'opt/openitc/etc/carbon' . DS . 'local_settings.py'    => '/opt/openitc/etc/carbon/local_settings.py',
            $this->configBase . 'docker' . DS . 'docker-compose.yml'                   => '/opt/openitc/docker/container/graphing/docker-compose.yml'
        ];

        foreach ($configFiles as $template => $outfile) {
            $this->io->out(sprintf('Generate %s', $outfile));

            $loader = new FilesystemLoader(dirname($template));
            $twig = new Environment($loader);

            file_put_contents($outfile, $twig->render(basename($template), $config));
        }
    }

    private function generateStatusengineBrokerModuleConfig() {
        $enableWebInterface = false;
        if (file_exists('/opt/openitc/etc/frontend/enable_web_interface')) {
            $enableWebInterface = true;
        }

        $config = [
            // storage-schemas.conf
            'ENABLE_FRONTEND_QUEUES' => 0
        ];

        if ($enableWebInterface === true) {
            // Enable required queues for the oITC Satellite Web Interface

            $config = [
                // storage-schemas.conf
                'ENABLE_FRONTEND_QUEUES' => 1
            ];

        }

        $template = $this->configBase . 'opt' . DS . 'openitc' . DS . 'etc' . DS . 'statusengine' . DS . 'statusengine.toml';
        $outfile = '/opt/openitc/etc/statusengine/statusengine.toml';
        $this->io->out(sprintf('Generate %s', $outfile));


        $loader = new FilesystemLoader(dirname($template));
        $twig = new Environment($loader);

        file_put_contents($outfile, $twig->render(basename($template), $config));
    }

}

