<?php
// Copyright (C) <2015-present>  <it-novum GmbH>
//
// This file is licensed under the terms of the openITCOCKPIT Enterprise Edition license agreement.
// The license agreement and license key were sent with the order confirmation.
declare(strict_types=1);

namespace App\Controller;

use App\Filter\GenericFilter;
use App\itnovum\openITCOCKPIT\Monitoring\Naemon\ExternalCommands;
use App\Model\Table\PushAgentsTable;
use Cake\Datasource\Exception\RecordNotFoundException;
use Cake\Http\Exception\BadRequestException;
use Cake\Http\Exception\MethodNotAllowedException;
use Cake\Http\Exception\NotFoundException;
use Cake\I18n\FrozenTime;
use Cake\Log\Log;
use itnovum\openITCOCKPIT\Core\ValueObjects\User;
use Cake\ORM\TableRegistry;
use itnovum\openITCOCKPIT\Database\PaginateOMat;

/**
 *
 * Class AgentconnectorController
 *
 *
 *
 * -------------------------> !!! READ THIS !!! <-------------------------
 * This file has to keep in sync with: https://github.com/it-novum/openITCOCKPIT/blob/development/src/Controller/AgentconnectorController.php
 * IF YOU MAKE A CHANGE IN THIS FILE MAKE SURE TO CHANGE BOTH
 * The files are slightly different on the Master and on the Satellite system - Please do not just copy & paste them over.
 *
 * This version only contains the required methods for openITCOCKPIT Monitoring Agents running in Push Mode
 */
class AgentconnectorController extends AppController {

    /*********************************
     *    AGENTS OVERVIEW METHODS    *
     *********************************/

    //Only for ACLs
    public function overview() {

    }

    public function pull() {
        if (!$this->isAngularJsRequest()) {
            //Only ship HTML Template
            return;
        }

        // On the satellite system, there is no AgentconfigsTable table
        // The satellite has no configuration for Agents that are running in PULL mode, because all required
        // configuration gets generated by the Master system.
        // However, we can read the generated configuration (JSON File), to get a list of all Pull agents.

        $pullAgents = [];

        $configFile = '/opt/openitc/receiver/etc/production.json';
        if (file_exists($configFile)) {
            $config = json_decode(file_get_contents($configFile), true);

            foreach ($config as $agentConfig) {
                if ($agentConfig['mode'] !== 'pull') {
                    continue;
                }

                // This rebuilds the API result from the Master System as close as possible
                $pullAgents[] = [
                    'host_id',
                    'port'          => $agentConfig['port'],
                    'use_https'     => $agentConfig['use_https'],
                    'insecure'      => $agentConfig['insecure'],
                    'use_autossl'   => $agentConfig['use_autossl'],
                    'use_push_mode' => true,
                    'basic_auth'    => $agentConfig['basic_auth'],
                    'host'          => [
                        'uuid'    => $agentConfig['uuid'],
                        'name'    => $agentConfig['name'],
                        'address' => $agentConfig['address'],
                    ]
                ];
            }
        }

        $this->set('agents', $pullAgents);
        $this->viewBuilder()->setOption('serialize', ['agents']);
    }

    public function push() {
        if (!$this->isAngularJsRequest()) {
            //Only ship HTML Template
            return;
        }

        /** @var PushAgentsTable $PushAgentsTable */
        $PushAgentsTable = TableRegistry::getTableLocator()->get('PushAgents');


        $GenericFilter = new GenericFilter($this->request);
        $GenericFilter->setFilters([
            'like' => [
                'HostMapping.display_name'
            ],
            'bool' => [
                'hasHostAssignment'
            ]
        ]);
        $PaginateOMat = new PaginateOMat($this, $this->isScrollRequest(), $GenericFilter->getPage());
        $User = new User($this->getUser());
        $UserTime = $User->getUserTime();


        $agents = $PushAgentsTable->getPushAgents($GenericFilter, $PaginateOMat);
        foreach ($agents as $index => $agent) {
            $agents[$index]['last_update'] = $UserTime->format($agent['last_update']);
        }

        $this->set('agents', $agents);
        $this->viewBuilder()->setOption('serialize', ['agents']);
    }

    public function showOutput() {
        if (!$this->isAngularJsRequest()) {
            //Only ship HTML Template
            return;
        }

        $isPullMode = $this->request->getQuery('mode', 'pull') === 'pull';
        $id = $this->request->getQuery('id', 0);

        if (empty($id) || $id <= 0) {
            throw new BadRequestException('ID is missing');
        }

        $checkresults = '[]';

        if ($isPullMode === true) {
            // It is not possible to query the Output of a Agent Running in Pull Mode on the Satellite System,
            // because the required certificate files are not readable by the Webserver.
            // I also think it is not a good idea the grant the webserver permissions to read the certificates
            // as this could cause major security issues!
            // For debugging or testing, you can query the Agent as user "nagios" from the CLI like so:
            //
            // /opt/openitc/receiver/bin/poller.php poller -H "fa0b2ddf-8c36-4e4e-bab9-9b5c867777b0" -c /opt/openitc/receiver/etc/production.json --dryrun

            throw new MethodNotAllowedException('Only Agents running in Push mode can be queried!');
        } else {
            // Agent is running in Push mode. Maybe has no config and host yet
            /** @var PushAgentsTable $PushAgentsTable */
            $PushAgentsTable = TableRegistry::getTableLocator()->get('PushAgents');

            $pushAgentId = $id;
            if (!$PushAgentsTable->existsById($pushAgentId)) {
                throw new NotFoundException();
            }

            $record = $PushAgentsTable->get($pushAgentId);

            $checkresults = $record->checkresults;
            if (empty($checkresults)) {
                $checkresults = '[]';
            }
        }

        $agent = [
            'uuid'                 => $record->uuid,
            'ipaddress'            => $record->ipaddress,
            'remote_address'       => $record->remote_address,
            'http_x_forwarded_for' => $record->http_x_forwarded_for
        ];

        $this->set('agent', $agent);
        $this->set('outputAsJson', $checkresults);
        $this->set('pushAgentUuid', $record->uuid);
        $this->viewBuilder()->setOption('serialize', ['outputAsJson', 'agent']);
    }

    /**
     * @param null $id
     */
    public function delete($id = null) {
        if (!$this->request->is('post')) {
            throw new MethodNotAllowedException();
        }

        /** @var PushAgentsTable $PushAgentsTable */
        $PushAgentsTable = TableRegistry::getTableLocator()->get('PushAgents');

        if (!$PushAgentsTable->existsById($id)) {
            throw new NotFoundException();
        }

        $entity = $PushAgentsTable->get($id);

        if ($PushAgentsTable->delete($entity)) {
            $this->set('success', true);
            $this->viewBuilder()->setOption('serialize', ['success']);

            return;
        }

        $this->response = $this->response->withStatus(400);
        $this->set('success', false);
        $this->viewBuilder()->setOption('serialize', ['success']);
        return;
    }

    /************************************
     *    AGENT API METHODS FOR PUSH    *
     ************************************/

    /**
     * Register new PUSH Agents
     *
     * How it works:
     * Register new Agents:
     * The Agent send it's UUID and an empty Password to the openITCOCKPIT Server. If no password was generated for the given UUID
     * openITCOCKPIT will generate a new Password and respond this password to the Agent. Respond with 201
     *
     * If an Agent sends an password which is not found in the database, openITCOCKPIT will respond with an 403 Forbidden
     */
    public function register_agent() {
        $agentUuid = $this->request->getData('agentuuid', null);
        $agentPassword = $this->request->getData('password', null);

        if (!$this->request->is('post') || !$this->isJsonRequest()) {
            throw new MethodNotAllowedException();
        }

        if ($agentUuid === null || $agentPassword === null) {
            $this->response = $this->response->withStatus(400);
            $this->set('error', 'Field uuid or password is missing in POST data');
            $this->viewBuilder()->setOption('serialize', ['error']);
            return;
        }

        /** @var PushAgentsTable $PushAgentsTable */
        $PushAgentsTable = TableRegistry::getTableLocator()->get('PushAgents');

        if ($agentPassword === '' && $PushAgentsTable->existsByUuid($agentUuid)) {
            // It this UUID already registered?
            $this->response = $this->response->withStatus(403);
            $this->set('error', 'The given UUID is already registed with a password!');
            $this->viewBuilder()->setOption('serialize', ['error']);
            return;
        }

        if ($agentPassword === '' && !$PushAgentsTable->existsByUuid($agentUuid)) {
            // New or unknown agent - Create a new password for this Agent and add it to our database
            $bytes = openssl_random_pseudo_bytes(64, $cstrong);
            $password = bin2hex($bytes);

            $remoteAddress = null;
            if (!empty($_SERVER['REMOTE_ADDR'])) {
                $remoteAddress = $_SERVER['REMOTE_ADDR'];
            }
            if (!empty($_SERVER['HTTP_CLIENT_IP'])) {
                $remoteAddress = $_SERVER['HTTP_CLIENT_IP'];
            }
            $HTTP_X_FORWARDED_FOR = null;
            if (!empty($_SERVER['HTTP_X_FORWARDED_FOR'])) {
                $HTTP_X_FORWARDED_FOR = $_SERVER['HTTP_X_FORWARDED_FOR'];
            }


            $entity = $PushAgentsTable->newEntity([
                'uuid'                 => $agentUuid,
                'host_uuid'            => null,
                'password'             => $password,
                'hostname'             => $this->request->getData('hostname', null),
                'ipaddress'            => $this->request->getData('ipaddress', null),
                'remote_address'       => $remoteAddress,
                'http_x_forwarded_for' => $HTTP_X_FORWARDED_FOR,
                'last_update'          => new FrozenTime(),
                'checkresults'         => null
            ]);

            $PushAgentsTable->save($entity);
            if ($entity->hasErrors()) {
                $this->response = $this->response->withStatus(400);
                $this->set('error', $entity->getErrors());
                $this->viewBuilder()->setOption('serialize', ['error']);
                return;
            }

            //Send new Password to Agent
            $this->response = $this->response->withStatus(201);
            $this->set('agentuuid', $agentUuid);
            $this->set('password', $password);
            $this->viewBuilder()->setOption('serialize', ['agentuuid', 'password']);
            return;
        }

        // Password and Agent UUID given - check if this exists in the database
        if ($PushAgentsTable->existsByUuidAndPassword($agentUuid, $agentPassword)) {
            $this->response = $this->response->withStatus(200);
            $this->set('success', true);
            $this->viewBuilder()->setOption('serialize', ['success']);
            return;
        }

        $this->response = $this->response->withStatus(404);
        $this->set('error', 'No Agent found for given UUID and password');
        $this->viewBuilder()->setOption('serialize', ['error']);
        return;
    }

    /**
     * Receiver function called by Agents running in PUSH mode
     */
    public function submit_checkdata() {
        if (!$this->isJsonRequest() || !$this->request->is('post')) {
            throw new MethodNotAllowedException();
        }

        $agentUuid = $this->request->getData('agentuuid', '');
        $agentPassword = $this->request->getData('password', '');
        $checkdata = $this->request->getData('checkdata', []);

        /** @var PushAgentsTable $PushAgentsTable */
        $PushAgentsTable = TableRegistry::getTableLocator()->get('PushAgents');

        $receivedChecks = 0;
        try {
            $pushAgent = $PushAgentsTable->getConfigWithHostForSubmitCheckdata(
                $agentUuid,
                $agentPassword
            );

            $hostUuid = $pushAgent->get('host_uuid');

            // Agent is known and authorized
            require_once "/opt/openitc/receiver/vendor/autoload.php";
            $CheckConfig = new \itnovum\openITCOCKPIT\Checks\Receiver\CheckConfig('/opt/openitc/receiver/etc/production.json');
            try {
                $config = $CheckConfig->getConfigByHostName($hostUuid);

                $results = [
                    'Messages' => []
                ];

                foreach ($config['checks'] as $pluginConfig) {
                    $pluginName = $pluginConfig['plugin'];

                    $pluginClassName = sprintf('itnovum\openITCOCKPIT\Checks\Receiver\Plugins\%s', $pluginName);
                    if (!class_exists($pluginClassName)) {
                        //Unknown Plugin
                        continue;
                    }

                    /** @var  $Plugin \itnovum\openITCOCKPIT\Checks\Receiver\Plugins\PluginInterface */
                    $Plugin = new $pluginClassName($pluginConfig, $checkdata);

                    $pluginOutput = $Plugin->getOutput();
                    if (strlen($Plugin->getPerfdataSerialized()) > 0) {
                        $pluginOutput .= '|' . $Plugin->getPerfdataSerialized();
                    }

                    // Create bulk message for Statusengine Broker
                    $results['messages'][] = [
                        'Command' => 'check_result',
                        'Data'    => [
                            'host_name'           => $hostUuid,
                            'service_description' => $pluginConfig['uuid'],
                            'output'              => $pluginOutput,
                            'long_output'         => ($Plugin->getLongOutput() === '' ? null : $Plugin->getLongOutput()),
                            'check_type'          => 1, //https://github.com/naemon/naemon-core/blob/cec6e10cbee9478de04b4cf5af29e83d47b5cfd9/src/naemon/common.h#L330-L334
                            'return_code'         => $Plugin->getStatuscode(),
                            'start_time'          => time(),
                            'end_time'            => time(),
                            'early_timeout'       => 0,
                            'latency'             => 0,
                            'exited_ok'           => 1
                        ]
                    ];

                    $receivedChecks++;
                }

                if (!empty($results['messages'])) {
                    $ExternalCommands = new ExternalCommands();
                    $ExternalCommands->toStatusnginCmdBackground($results);
                }

            } catch (\RuntimeException $e) {
                // Host was not exported yet to Monitoring Engine and check receiver - just save the check result to the database
                // and ignore the error
            }

            $pushAgent->set('last_update', new FrozenTime());
            $pushAgent->set('checkresults', json_encode($checkdata));
            $PushAgentsTable->save($pushAgent);

            $this->set('success', true);
            $this->set('received_checks', $receivedChecks);
            $this->viewBuilder()->setOption('serialize', ['success', 'received_checks']);
            return;
        } catch (RecordNotFoundException $e) {
            //Agent has no host_uuid assignment yet. Due to we are on the satellite, we store the check results anyway
            //so we can answer to possible requests from the master system.

            try {
                $pushAgent = $PushAgentsTable->getRecordByAgentUuidAndPassword($agentUuid, $agentPassword);
                $pushAgent->set('last_update', new FrozenTime());
                $pushAgent->set('checkresults', json_encode($checkdata));
                $PushAgentsTable->save($pushAgent);
            } catch (RecordNotFoundException $e) {
                // Unknown agent - we dont save anything
            }
        }

        $this->response = $this->response->withStatus(400);
        $this->set('error', 'Invalid credentials or host not found');
        $this->viewBuilder()->setOption('serialize', ['error']);
    }
}
