<?php
// Copyright (C) <2015-present>  <it-novum GmbH>
//
// This file is licensed under the terms of the openITCOCKPIT Enterprise Edition license agreement.
// The license agreement and license key were sent with the order confirmation.
declare(strict_types=1);

namespace App\Controller;

use App\Filter\UserFilter;
use App\Lib\Api;
use App\Lib\SystemConfig;
use App\Model\Entity\User;
use App\Model\Table\UsersTable;
use Authentication\Authenticator\ResultInterface;
use Authentication\Controller\Component\AuthenticationComponent;
use Cake\Cache\Cache;
use Cake\Event\EventInterface;
use Cake\Http\Exception\MethodNotAllowedException;
use Cake\Http\Exception\NotFoundException;
use Cake\ORM\TableRegistry;
use itnovum\openITCOCKPIT\Core\LoginBackgrounds;
use itnovum\openITCOCKPIT\Core\Views\UserTime;
use itnovum\openITCOCKPIT\Database\PaginateOMat;

/**
 * Class UsersController
 * @package App\Controller
 *
 * @property AuthenticationComponent $Authentication
 */
class UsersController extends AppController {

    public function beforeFilter(EventInterface $event) {
        parent::beforeFilter($event);
        $this->Authentication->allowUnauthenticated(['login']);
    }

    public function index() {
        if (!$this->isApiRequest()) {
            //Only ship the template for .html requests
            return;
        }

        /** @var UsersTable $UsersTable */
        $UsersTable = TableRegistry::getTableLocator()->get('Users');

        $UserFilter = new UserFilter($this->request);
        $PaginateOMat = new PaginateOMat($this, $this->isScrollRequest(), $UserFilter->getPage());

        $users = $UsersTable->getUsersByFilterRequest($UserFilter, $PaginateOMat);

        $this->set('users', $users);
        $this->set('myUserId', $this->getUser()->get('id'));

        $this->viewBuilder()->setOption('serialize', ['users', 'myUserId']);
    }

    public function add() {
        if (!$this->isApiRequest()) {
            //Only ship HTML template for angular
            return;
        }

        /** @var UsersTable $UsersTable */
        $UsersTable = TableRegistry::getTableLocator()->get('Users');
        if ($this->request->is('post') || $this->request->is('put')) {

            $data = $this->request->getData('User', []);
            $user = $UsersTable->newEmptyEntity();
            $user = $UsersTable->patchEntity($user, $data);

            $UsersTable->save($user);
            if ($user->hasErrors()) {
                $this->response = $this->response->withStatus(400);
                $this->set('error', $user->getErrors());
                $this->viewBuilder()->setOption('serialize', ['error']);
                return;
            }
            Cache::clear('permissions');
            $this->set('user', $user);
            $this->viewBuilder()->setOption('serialize', ['user']);
        }
    }

    /**
     * @param int|null $id
     */
    public function edit($id = null) {
        if (!$this->isApiRequest()) {
            //Only ship HTML template for angular
            return;
        }

        /** @var UsersTable $UsersTable */
        $UsersTable = TableRegistry::getTableLocator()->get('Users');

        if (!$UsersTable->existsById($id)) {
            throw new NotFoundException(__('User not found'));
        }

        if ($this->request->is('get')) {
            $user = $UsersTable->get($id);
            $this->set('user', $user);
            $this->viewBuilder()->setOption('serialize', ['user']);
            return;
        }

        if ($this->request->is(['patch', 'post', 'put'])) {
            $data = $this->request->getData('User', []);

            //prevent multiple hash of password
            if ($data['password'] === '' && $data['confirm_password'] === '') {
                unset($data['password']);
                unset($data['confirm_password']);
            }

            /** @var User $user */
            $user = $UsersTable->get($id);
            $user = $UsersTable->patchEntity($user, $data);

            $UsersTable->save($user);
            if ($user->hasErrors()) {
                $this->response = $this->response->withStatus(400);
                $this->set('error', $user->getErrors());
                $this->viewBuilder()->setOption('serialize', ['error']);
                return;
            }
            Cache::clear('permissions');
            $this->set('user', $user);
            $this->viewBuilder()->setOption('serialize', ['user']);
        }
    }


    public function delete($id = null) {
        if (!$this->request->is('post') || !$this->isApiRequest()) {
            throw new MethodNotAllowedException();
        }

        /** @var UsersTable $UsersTable */
        $UsersTable = TableRegistry::getTableLocator()->get('Users');

        if (!$UsersTable->existsById($id)) {
            throw new NotFoundException(__('User not found'));
        }

        if ($id == $this->getUser()->get('id')) {
            throw new \RuntimeException('You cannot delete yourself!');
        }

        $user = $UsersTable->get($id);
        if ($UsersTable->delete($user)) {
            $this->set('success', true);
            $this->viewBuilder()->setOption('serialize', ['success']);

            return;
        }

        $this->response = $this->response->withStatus(400);
        $this->set('success', false);
        $this->viewBuilder()->setOption('serialize', ['success']);
        return;
    }

    public function login() {
        $this->viewBuilder()->setLayout('login');
        $LoginBackgrounds = new LoginBackgrounds();
        $images = $LoginBackgrounds->getImages();
        $SystemConfig = new SystemConfig();

        $disableAnimation = $this->request->getQuery('remote', null) !== null;

        if ($SystemConfig->getDisableLoginAnimation()) {
            $disableAnimation = true;
        }

        if ($disableAnimation) {
            $images['particles'] = 'none';
        }

        $hasValidSslCertificate = false;
        if (isset($_SERVER['SSL_VERIFIED']) && $_SERVER['SSL_VERIFIED'] === 'SUCCESS' && isset($_SERVER['SSL_CERT'])) {
            $hasValidSslCertificate = $this->getUser() !== null;
        }

        $this->set('_csrfToken', $this->request->getParam('_csrfToken'));
        $this->set('images', $images);
        $this->set('disableAnimation', $disableAnimation);
        $this->set('hasValidSslCertificate', $hasValidSslCertificate);

        if ($this->request->is('get')) {
            $this->viewBuilder()->setOption('serialize', ['_csrfToken', 'images', 'hasValidSslCertificate']);
            return;
        }

        /** @var UsersTable $UsersTable */
        $UsersTable = TableRegistry::getTableLocator()->get('Users');

        if ($this->request->is('post')) {
            // Docs: https://book.cakephp.org/authentication/1/en/index.html

            $result = $this->Authentication->getResult();
            if ($result->getStatus() === ResultInterface::SUCCESS) {
                $this->set('success', true);
                $this->viewBuilder()->setOption('serialize', ['success']);
                return;
            }

            $this->RequestHandler->renderAs($this, 'json');
            $this->response = $this->response->withStatus(400);
            $errors = $result->getErrors();
            $errors['Password'][] = __('Invalid username or password');

            $this->set('success', false);
            $this->set('errors', $errors);
            $this->viewBuilder()->setOption('serialize', ['success', 'errors']);
        }
    }

    public function logout() {
        $this->Authentication->logout();

        $this->redirect([
            'action' => 'login'
        ]);
    }

    /****************************
     *       AJAX METHODS       *
     ****************************/

    public function loadUsergroups() {
        if (!$this->isApiRequest()) {
            throw new MethodNotAllowedException();
        }

        /** @var UsergroupsTable $UsergroupsTable */
        $UsergroupsTable = TableRegistry::getTableLocator()->get('Usergroups');

        $_usergroups = $UsergroupsTable->find('list')->toArray();

        //Rewrite hashmap to array because javascript will break the "orderd hashmap" because hashmaps have no order
        $usergroups = [];
        foreach ($_usergroups as $id => $name) {
            $usergroups[] = [
                'key'   => $id,
                'value' => $name
            ];
        }

        $this->set('usergroups', $usergroups);
        $this->viewBuilder()->setOption('serialize', ['usergroups']);
    }

    public function loadDateformats() {
        if (!$this->isApiRequest()) {
            //Only ship HTML template for angular
            return;
        }
        /** @var UsersTable $UsersTable */
        $UsersTable = TableRegistry::getTableLocator()->get('Users');
        $dateformats = $UsersTable->getDateformats();
        $options = [];

        $User = $this->getUser();
        foreach ($dateformats as $dateformat) {
            $UserTime = new UserTime($User->get('timezone'), $dateformat);
            $options[$dateformat] = $UserTime->format(time());
        }
        $dateformats = Api::makeItJavaScriptAble($options);
        $defaultDateFormat = 'H:i:s - d.m.Y'; // key 10

        $this->set('dateformats', $dateformats);
        $this->set('defaultDateFormat', $defaultDateFormat);
        $this->viewBuilder()->setOption('serialize', ['dateformats', 'defaultDateFormat']);
    }
}
