<?php
// Copyright (C) <2015-present>  <it-novum GmbH>
//
// This file is licensed under the terms of the openITCOCKPIT Enterprise Edition license agreement.
// The license agreement and license key were sent with the order confirmation.

declare(strict_types=1);


namespace App\Model\Table;

use App\Lib\Traits\PaginationAndScrollIndexTrait;
use App\Model\Entity\DowntimeHost;
use Cake\ORM\RulesChecker;
use Cake\ORM\Table;
use Cake\Utility\Hash;
use Cake\Validation\Validator;
use itnovum\openITCOCKPIT\Core\DowntimeServiceConditions;
use itnovum\openITCOCKPIT\Database\PaginateOMat;

/**
 * DowntimeService Model
 *
 * @method \Model\Entity\DowntimeService newEmptyEntity()
 * @method \Model\Entity\DowntimeService newEntity(array $data, array $options = [])
 * @method \Model\Entity\DowntimeService[] newEntities(array $data, array $options = [])
 * @method \Model\Entity\DowntimeService get($primaryKey, $options = [])
 * @method \Model\Entity\DowntimeService findOrCreate($search, ?callable $callback = null, $options = [])
 * @method \Model\Entity\DowntimeService patchEntity(\Cake\Datasource\EntityInterface $entity, array $data, array $options = [])
 * @method \Model\Entity\DowntimeService[] patchEntities(iterable $entities, array $data, array $options = [])
 * @method \Model\Entity\DowntimeService|false save(\Cake\Datasource\EntityInterface $entity, $options = [])
 * @method \Model\Entity\DowntimeService saveOrFail(\Cake\Datasource\EntityInterface $entity, $options = [])
 * @method \Model\Entity\DowntimeService[]|\Cake\Datasource\ResultSetInterface|false saveMany(iterable $entities, $options = [])
 * @method \Model\Entity\DowntimeService[]|\Cake\Datasource\ResultSetInterface saveManyOrFail(iterable $entities, $options = [])
 * @method \Model\Entity\DowntimeService[]|\Cake\Datasource\ResultSetInterface|false deleteMany(iterable $entities, $options = [])
 * @method \Model\Entity\DowntimeService[]|\Cake\Datasource\ResultSetInterface deleteManyOrFail(iterable $entities, $options = [])
 */
class DowntimeServicesTable extends Table {

    use PaginationAndScrollIndexTrait;

    /**
     * Initialize method
     *
     * @param array $config The configuration for the Table.
     * @return void
     */
    public function initialize(array $config): void {
        parent::initialize($config);

        $this->setTable('statusengine_service_downtimehistory');
        $this->setDisplayField('hostname');
        $this->setPrimaryKey(['hostname', 'service_description', 'node_name', 'scheduled_start_time', 'internal_downtime_id']);
    }

    /**
     * Default validation rules.
     *
     * @param \Cake\Validation\Validator $validator Validator instance.
     * @return \Cake\Validation\Validator
     */
    public function validationDefault(Validator $validator): Validator {
        //Readonly table
        return $validator;
    }

    /**
     * Returns a rules checker object that will be used for validating
     * application integrity.
     *
     * @param \Cake\ORM\RulesChecker $rules The rules object to be modified.
     * @return \Cake\ORM\RulesChecker
     */
    public function buildRules(RulesChecker $rules): RulesChecker {
        //Readonly table
        return $rules;
    }

    /**
     * @param DowntimeServiceConditions $DowntimeServiceConditions
     * @param PaginateOMat|null $PaginateOMat
     * @return array|void
     */
    public function getDowntimes(DowntimeServiceConditions $DowntimeServiceConditions, $PaginateOMat = null) {
        $query = $this->find();
        $query->select([
            'DowntimeServices.hostname',
            'DowntimeServices.service_description',
            'DowntimeServices.author_name',
            'DowntimeServices.comment_data',
            'DowntimeServices.entry_time',
            'DowntimeServices.scheduled_start_time',
            'DowntimeServices.scheduled_end_time',
            'DowntimeServices.duration',
            'DowntimeServices.was_started',
            'DowntimeServices.internal_downtime_id',
            'DowntimeServices.was_cancelled'
        ])
            ->where([
                'DowntimeServices.scheduled_start_time >' => $DowntimeServiceConditions->getFrom(),
                'DowntimeServices.scheduled_start_time <' => $DowntimeServiceConditions->getTo(),
            ])
            ->order($DowntimeServiceConditions->getOrder());


        if ($DowntimeServiceConditions->hideExpired()) {
            $query->andWhere([
                'DowntimeServices.scheduled_end_time >' => time()
            ]);
        }

        if ($DowntimeServiceConditions->hasConditions()) {

            $where = $DowntimeServiceConditions->getConditions();

            if (!empty($where))
                $query->andWhere($where);

        }

        if ($DowntimeServiceConditions->isRunning()) {
            $query->andWhere([
                'DowntimeServices.scheduled_end_time >' => time(),
                'DowntimeServices.was_started'          => 1,
                'DowntimeServices.was_cancelled'        => 0
            ]);
        }

        if ($PaginateOMat === null) {
            //Just execute query
            $result = $this->emptyArrayIfNull($query->toArray());
        } else {
            if ($PaginateOMat->useScroll()) {
                $result = $this->scrollCake4($query, $PaginateOMat->getHandler());
            } else {
                $result = $this->paginateCake4($query, $PaginateOMat->getHandler());
            }
        }

        return $result;
    }


    /**
     * @param null $uuid
     * @param bool $isRunning
     * @return array|\Cake\Datasource\EntityInterface|null
     */
    public function byServiceUuid($uuid = null, $isRunning = false) {
        if (empty($uuid)) {
            return null;
        }

        $query = $this->find();
        $query->select([
            'DowntimeServices.hostname',
            'DowntimeServices.service_description',
            'DowntimeServices.author_name',
            'DowntimeServices.comment_data',
            'DowntimeServices.entry_time',
            'DowntimeServices.scheduled_start_time',
            'DowntimeServices.scheduled_end_time',
            'DowntimeServices.duration',
            'DowntimeServices.was_started',
            'DowntimeServices.internal_downtime_id',
            'DowntimeServices.was_cancelled',
        ])
            ->order([
                'DowntimeServices.entry_time' => 'DESC'
            ])
            ->where([
                'DowntimeServices.service_description' => $uuid
            ]);

        if ($isRunning) {
            $query->andWhere([
                'DowntimeServices.scheduled_end_time >' => time(),
                'DowntimeServices.was_started'          => 1,
                'DowntimeServices.was_cancelled'        => 0

            ]);
        }

        return $query->first();
    }

    /**
     * @param string $hostUuid
     * @param array $DowntimeHost
     * @return array|\ArrayAccess
     */
    public function getServiceDowntimesByHostUuidAndDowntime(string $hostUuid, array $downtimeHost) {
        $records = $this->find()
            ->where([
                'DowntimeServices.hostname'             => $hostUuid,
                'DowntimeServices.scheduled_start_time' => $downtimeHost['scheduled_start_time'],
                'DowntimeServices.scheduled_end_time'   => $downtimeHost['scheduled_end_time']
            ])
            ->disableHydration()
            ->all();

        if ($records === null) {
            return [];
        }

        return $records->toArray();
    }

    /**
     * @param int $internalDowntimeId
     * @return array
     */
    public function getServiceUuidWithDowntimeByInternalDowntimeId($internalDowntimeId) {
        $result = $this->find()
            ->select([
                'DowntimeServices.hostname',
                'DowntimeServices.service_description',
                'DowntimeServices.author_name',
                'DowntimeServices.comment_data',
                'DowntimeServices.entry_time',
                'DowntimeServices.scheduled_start_time',
                'DowntimeServices.scheduled_end_time',
                'DowntimeServices.duration',
                'DowntimeServices.was_started',
                'DowntimeServices.internal_downtime_id',
                'DowntimeServices.was_cancelled'
            ])
            ->where([
                'DowntimeServices.internal_downtime_id' => $internalDowntimeId
            ])
            ->disableHydration()
            ->first();

        if ($result === null) {
            return [];
        }

        return $result;
    }
}
