<?php
// Copyright (C) <2015-present>  <it-novum GmbH>
//
// This file is licensed under the terms of the openITCOCKPIT Enterprise Edition license agreement.
// The license agreement and license key were sent with the order confirmation.

namespace App\Model\Table;

use App\Filter\HoststatusFilter;
use App\Lib\Traits\PaginationAndScrollIndexTrait;
use App\Model\Entity\Hoststatus;
use Cake\Datasource\EntityInterface;
use Cake\ORM\RulesChecker;
use Cake\ORM\Table;
use Cake\Validation\Validator;
use itnovum\openITCOCKPIT\Database\PaginateOMat;

/**
 * Hoststatus Model
 *
 * @method Hoststatus get($primaryKey, $options = [])
 * @method Hoststatus newEntity($data = null, array $options = [])
 * @method Hoststatus[] newEntities(array $data, array $options = [])
 * @method Hoststatus|bool save(EntityInterface $entity, $options = [])
 * @method Hoststatus patchEntity(EntityInterface $entity, array $data, array $options = [])
 * @method Hoststatus[] patchEntities($entities, array $data, array $options = [])
 * @method Hoststatus findOrCreate($search, callable $callback = null, $options = [])
 */
class HoststatusTable extends Table {
    use PaginationAndScrollIndexTrait;

    /**
     * Initialize method
     *
     * @param array $config The configuration for the Table.
     * @return void
     */
    public function initialize(array $config): void {
        parent::initialize($config);

        $this->setTable('statusengine_hoststatus');
        $this->setDisplayField('hostname');
        $this->setPrimaryKey('hostname');

        $this->hasOne('HostMapping', [
            'className'  => 'Mapping',
            'foreignKey' => 'uuid',
            'joinType'   => 'INNER'
        ]);

        $this->hasOne('TimeperiodMapping', [
            'className'  => 'Mapping',
            'foreignKey' => 'uuid',
            'bindingKey' => 'check_timeperiod',
            'joinType'   => 'INNER'
        ]);
    }

    /**
     * Default validation rules.
     *
     * @param \Cake\Validation\Validator $validator Validator instance.
     * @return \Cake\Validation\Validator
     */
    public function validationDefault(Validator $validator): Validator {
        //Readonly table
        return $validator;
    }

    /**
     * Returns a rules checker object that will be used for validating
     * application integrity.
     *
     * @param \Cake\ORM\RulesChecker $rules The rules object to be modified.
     * @return \Cake\ORM\RulesChecker
     */
    public function buildRules(RulesChecker $rules): RulesChecker {
        //Readonly table
        return $rules;
    }

    /**
     * @param bool $includeOkState
     * @return array
     */
    public function getHoststatusCountStatusengine3($includeOkState = false) {
        $hoststatusCount = [
            '1' => 0,
            '2' => 0,
        ];
        if ($includeOkState === true) {
            $hoststatusCount['0'] = 0;
        }

        $query = $this->find();
        $query
            ->select([
                'current_state',
                'count' => $query->newExpr('COUNT(current_state)'),
            ])
            ->group([
                'current_state',
            ])
            ->disableHydration();

        if ($includeOkState === false) {
            $query->andWhere([
                'current_state >' => 0
            ]);
        }

        $hoststatusCountResult = $query->all();

        foreach ($hoststatusCountResult as $hoststatus) {
            $hoststatusCount[$hoststatus['current_state']] = (int)$hoststatus['count'];
        }
        return $hoststatusCount;
    }

    /**
     * @param HoststatusFilter $HoststatusFilter
     * @param PaginateOMat|null $PaginateOMat
     */
    public function getHoststatusIndex(HoststatusFilter $HoststatusFilter, ?PaginateOMat $PaginateOMat = null) {
        $query = $this->find();
        $query->contain([
            'HostMapping'
        ]);
        $query->where($HoststatusFilter->indexFilter());
        $query->order($HoststatusFilter->getOrderForPaginator('current_state', 'desc'));


        if ($PaginateOMat === null) {
            //Just execute query
            $result = $query->toArray();
        } else {
            if ($PaginateOMat->useScroll()) {
                $result = $this->scrollCake4($query, $PaginateOMat->getHandler());
            } else {
                $result = $this->paginateCake4($query, $PaginateOMat->getHandler());
            }
        }
        return $result;
    }

    /**
     * @param string $uuid
     * @return Hoststatus|null
     */
    public function getHoststatusDetails(string $uuid): ?Hoststatus {
        $hoststatus = $this->find()
            ->contain([
                'HostMapping',
                'TimeperiodMapping'
            ])
            ->where(['hostname' => $uuid])
            ->first();

        return $hoststatus;
    }

    /**
     * @return int
     */
    public function getHoststatusCountForDashboard() {
        return $this->find()->count();
    }

    /**
     * @return array
     */
    public function getHoststateSummary() {
        $hostSummaryState = [
            'up'          => 0,
            'down'        => 0,
            'unreachable' => 0
        ];

        $query = $this->find();
        $query->select([
            'current_state',
            'count' => $query->func()->count('*')
        ])
            ->group('current_state');

        $Hoststatus = new \App\Hoststatus();
        foreach ($query as $record) {
            $hostSummaryState[$Hoststatus->getStatenameByStateId($record->current_state)] = $record->count;
        }

        return $hostSummaryState;
    }

    /**
     * @param $hoststatus
     * @param bool $extended show details ('acknowledged', 'in downtime', ...)
     * @return array
     */
    public function getHostStateSummaryFromHoststatus($hoststatus, $extended = true) {
        $hostStateSummary = [
            'state' => [
                0 => 0,
                1 => 0,
                2 => 0
            ],
            'total' => 0
        ];
        if ($extended === true) {
            $hostStateSummary = [
                'state'        => [
                    0 => 0,
                    1 => 0,
                    2 => 0
                ],
                'acknowledged' => [
                    0 => 0,
                    1 => 0,
                    2 => 0
                ],
                'in_downtime'  => [
                    0 => 0,
                    1 => 0,
                    2 => 0
                ],
                'not_handled'  => [
                    0 => 0,
                    1 => 0,
                    2 => 0
                ],
                'passive'      => [
                    0 => 0,
                    1 => 0,
                    2 => 0
                ],
                'total'        => 0
            ];
        }
        if (empty($hoststatus)) {
            return $hostStateSummary;
        }

        foreach ($hoststatus as $hostUuid => $status) {
            //Check for randome exit codes like 255...
            if ($status['current_state'] > 2) {
                $status['current_state'] = 2;
            }

            $hostStateSummary['state'][$status['current_state']]++;
            if ($extended === true) {
                if ($status['current_state'] > 0) {
                    if ($status['problem_has_been_acknowledged'] > 0) {
                        $hostStateSummary['acknowledged'][$status['current_state']]++;
                    } else {
                        $hostStateSummary['not_handled'][$status['current_state']]++;
                    }
                }

                if ($status['scheduled_downtime_depth'] > 0) {
                    $hostStateSummary['in_downtime'][$status['current_state']]++;
                }
                if ($status['active_checks_enabled'] == 0) {
                    $hostStateSummary['passive'][$status['current_state']]++;
                }
            }
            $hostStateSummary['total']++;
        }
        return $hostStateSummary;
    }

    public function byUuid($uuid) {
        $query = $this->find()
            ->where([
                'Hoststatus.hostname' => $uuid
            ])
            ->first();

        return $query;
    }

    /**
     * @param $uuids
     * @param bool $enableHydration
     * @return array
     */
    public function byUuids($uuids, $enableHydration = false) {
        if (!is_array($uuids)) {
            $uuids = [$uuids];
        }

        $query = $this->find()
            ->where([
                'Hoststatus.hostname IN' => $uuids
            ])
            ->enableHydration($enableHydration)
            ->all();

        if (empty($query)) {
            return [];
        }

        $hoststatus = [];
        foreach ($query->toArray() as $record) {
            $hoststatus[$record['hostname']] = $record;
        }

        return $hoststatus;
    }

    /**
     * @param string $uuid
     * @return bool
     */
    public function existsByUuid(?string $uuid) {
        if ($uuid === null) {
            return false;
        }
        return $this->exists(['Hoststatus.hostname' => $uuid]);
    }

    /**
     * @param $hoststatus
     * @param bool $extended show details ('acknowledged', 'in downtime', ...)
     * @return array
     */
    public function getHostStateSummaryForMaps($hoststatus, $extended = true) {
        $hostStateSummary = [
            'state' => [
                0 => 0,
                1 => 0,
                2 => 0
            ],
            'total' => 0
        ];
        if ($extended === true) {
            $hostStateSummary = [
                'state'        => [
                    0 => 0,
                    1 => 0,
                    2 => 0
                ],
                'acknowledged' => [
                    0 => 0,
                    1 => 0,
                    2 => 0
                ],
                'in_downtime'  => [
                    0 => 0,
                    1 => 0,
                    2 => 0
                ],
                'not_handled'  => [
                    0 => 0,
                    1 => 0,
                    2 => 0
                ],
                'passive'      => [
                    0 => 0,
                    1 => 0,
                    2 => 0
                ],
                'total'        => 0
            ];
        }
        if (empty($hoststatus)) {
            return $hostStateSummary;
        }
        foreach ($hoststatus as $host) {
            //Check for randome exit codes like 255...
            if ($host['current_state'] > 2) {
                $host['current_state'] = 2;
            }

            $hostStateSummary['state'][$host['current_state']]++;
            if ($extended === true) {
                if ($host['current_state'] > 0) {
                    if ($host['problem_has_been_acknowledged'] > 0) {
                        $hostStateSummary['acknowledged'][$host['current_state']]++;
                    } else {
                        $hostStateSummary['not_handled'][$host['current_state']]++;
                    }
                }

                if ($host['scheduled_downtime_depth'] > 0) {
                    $hostStateSummary['in_downtime'][$host['current_state']]++;
                }
                if ($host['active_checks_enabled'] == 0) {
                    $hostStateSummary['passive'][$host['current_state']]++;
                }
            }
            $hostStateSummary['total']++;
        }
        return $hostStateSummary;
    }

    /**
     * @param $uuids
     * @param bool $enableHydration
     * @return array
     */
    public function byUuidsWithMapping($uuids, $enableHydration = false) {
        if (!is_array($uuids)) {
            $uuids = [$uuids];
        }

        $query = $this->find()
            ->where([
                'Hoststatus.hostname IN' => $uuids
            ])
            ->contain([
                'HostMapping'
            ])
            ->enableHydration($enableHydration)
            ->all();

        if (empty($query)) {
            return [];
        }

        $hoststatus = [];
        foreach ($query->toArray() as $record) {
            $hoststatus[$record['hostname']] = $record;
        }

        return $hoststatus;
    }

    /**
     * @param string $uuid
     * @param bool $enableHydration
     * @return array
     */
    public function byUuidWithMapping($uuid, $enableHydration = false) {
        $query = $this->find()
            ->where([
                'Hoststatus.hostname' => $uuid
            ])
            ->contain([
                'HostMapping'
            ])
            ->enableHydration($enableHydration)
            ->first();

        if (empty($query)) {
            return [];
        }

        return $query;
    }
}
