<?php
// Copyright (C) <2015-present>  <it-novum GmbH>
//
// This file is licensed under the terms of the openITCOCKPIT Enterprise Edition license agreement.
// The license agreement and license key were sent with the order confirmation.

namespace App\Model\Table;

use App\Filter\UserFilter;
use App\Lib\Traits\PaginationAndScrollIndexTrait;
use App\Model\Entity\User;
use Authentication\PasswordHasher\DefaultPasswordHasher;
use Authentication\PasswordHasher\PasswordHasherFactory;
use Authentication\PasswordHasher\PasswordHasherInterface;
use Authentication\PasswordHasher\PasswordHasherTrait;
use Cake\Datasource\EntityInterface;
use Cake\Event\Event;
use Cake\ORM\Behavior\TimestampBehavior;
use Cake\ORM\RulesChecker;
use Cake\ORM\Table;
use Cake\Validation\Validator;
use itnovum\openITCOCKPIT\Core\UUID;
use itnovum\openITCOCKPIT\Database\PaginateOMat;

/**
 * Users Model
 *
 * @property \App\Model\Table\UsergroupsTable|\Cake\ORM\Association\BelongsTo $Usergroups
 * @property \App\Model\Table\ApikeysTable|\Cake\ORM\Association\HasMany $Apikeys
 *
 * @method User get($primaryKey, $options = [])
 * @method User newEntity($data = null, array $options = [])
 * @method User[] newEntities(array $data, array $options = [])
 * @method User|bool save(EntityInterface $entity, $options = [])
 * @method User patchEntity(EntityInterface $entity, array $data, array $options = [])
 * @method User[] patchEntities($entities, array $data, array $options = [])
 * @method User findOrCreate($search, callable $callback = null, $options = [])
 *
 * @mixin TimestampBehavior
 */
class UsersTable extends Table {

    use PaginationAndScrollIndexTrait;

    use PasswordHasherTrait {
        getPasswordHasher as protected _getPasswordHasher;
    }

    /**
     * Password validation regex.
     */
    const PASSWORD_REGEX = '/^(?=.*\d).{6,}$/i';

    /**
     * Initialize method
     *
     * @param array $config The configuration for the Table.
     * @return void
     */
    public function initialize(array $config): void {
        parent::initialize($config);

        $this->setTable('users');
        $this->setDisplayField('id');
        $this->setPrimaryKey('id');

        $this->belongsTo('Usergroups', [
            'foreignKey' => 'usergroup_id',
            'joinType'   => 'INNER'
        ]);

        $this->hasMany('Apikeys', [
            'foreignKey' => 'user_id'
        ]);

        $this->addBehavior('Timestamp');
    }

    /**
     * Default validation rules.
     *
     * @param Validator $validator Validator instance.
     * @return Validator
     */
    public function validationDefault(Validator $validator): Validator {
        $validator
            ->integer('id')
            ->allowEmptyString('id', null, 'create');

        $validator
            ->integer('usergroup_id')
            ->requirePresence('usergroup_id', 'create')
            ->greaterThan('usergroup_id', 0, __('You have to select a user role.'))
            ->allowEmptyString('usergroup_id', null, false);

        $validator
            ->requirePresence('username', 'create')
            ->notEmptyString('username');

        $validator
            ->scalar('password')
            ->requirePresence('password', 'create')
            ->allowEmptyString('password', null, false)
            ->regex('password', self::PASSWORD_REGEX, 'The password must consist of 6 alphanumeric characters and must contain at least one digit.');

        $validator->add('confirm_password',
            'compareWith', [
                'rule'    => ['compareWith', 'password'],
                'message' => 'Passwords not equal'
            ]);

        $validator
            ->scalar('firstname')
            ->maxLength('firstname', 100)
            ->requirePresence('firstname', 'create')
            ->allowEmptyString('firstname', null, false);

        $validator
            ->scalar('lastname')
            ->maxLength('lastname', 100)
            ->requirePresence('lastname', 'create')
            ->allowEmptyString('lastname', null, false);

        return $validator;
    }

    /**
     * Returns a rules checker object that will be used for validating
     * application integrity.
     *
     * @param RulesChecker $rules The rules object to be modified.
     * @return RulesChecker
     */
    public function buildRules(RulesChecker $rules): RulesChecker {
        $rules->add($rules->isUnique(['username']));
        $rules->add($rules->existsIn(['usergroup_id'], 'Usergroups'));

        return $rules;
    }

    /**
     * @param Event $event
     * @param EntityInterface $entity
     * @param \ArrayObject $options
     * @return bool
     */
    public function beforeSave(Event $event, EntityInterface $entity, \ArrayObject $options) {
        if ($entity->isDirty('password')) {
            $Hasher = $this->getPasswordHasher();
            $entity->password = $Hasher->hash($entity->password);
        }
        return true;
    }

    /**
     * Return password hasher object.
     *
     * @return PasswordHasherInterface Password hasher instance.
     */
    public function getPasswordHasher(): PasswordHasherInterface {
        return $this->_getPasswordHasher();
    }

    /**
     * @param string $username
     * @return array|EntityInterface|null
     */
    public function getUserByUsernameForLogin(string $username) {
        $query = $this->find();
        return $query
            ->where([
                'Users.username' => $username
            ])
            ->first();
    }

    /**
     * @param string $apikey
     * @return array|EntityInterface|null
     */
    public function getUserByApikeyForLogin(string $apikey) {
        $query = $this->find()
            ->join([
                [
                    'table'      => 'apikeys',
                    'alias'      => 'Apikeys',
                    'type'       => 'INNER',
                    'conditions' => [
                        'Apikeys.user_id = Users.id'
                    ]
                ]
            ])
            ->enableAutoFields()
            ->where([
                'Apikeys.apikey' => $apikey
            ]);
        return $query->first();
    }


    /**
     * @param $id
     * @return array|EntityInterface|null
     */
    public function getUserById($id) {
        $query = $this->find()
            ->disableHydration()
            ->where([
                'Users.id' => $id
            ]);
        if (is_null($query)) {
            return $this->newEmptyEntity();
        }
        return $query->first();
    }

    public function getUsersByFilterRequest(UserFilter $UserFilter, PaginateOMat $PaginateOMat) {
        $query = $this->find()
            ->contain([
                'Usergroups'
            ])
            ->where($UserFilter->indexFilter())
            ->order($UserFilter->getOrderForPaginator('Users.username', 'asc'));

        if ($PaginateOMat === null) {
            //Just execute query
            $result = $query->toArray();
        } else {
            if ($PaginateOMat->useScroll()) {
                $result = $this->scrollCake4($query, $PaginateOMat->getHandler());
            } else {
                $result = $this->paginateCake4($query, $PaginateOMat->getHandler());
            }
        }

        return $result;
    }

    /**
     * @param int $id
     * @return bool
     */
    public function existsById($id) {
        return $this->exists(['Users.id' => $id]);
    }

    /**
     * @return array
     */
    public function getDateformats() {
        //Format: https://www.php.net/manual/de/function.date.php
        return [
            1 => 'F j, Y H:i:s',        //'%B %e, %Y %H:%M:%S',
            2 => 'm-d-Y H:i:s',         //'%m-%d-%Y  %H:%M:%S',
            3 => 'm-d-Y H:i',           //'%m-%d-%Y  %H:%M',
            4 => 'm-d-Y h:i:s A',       //'%m-%d-%Y  %l:%M:%S %p',
            5 => 'H:i:s m-d-Y',         //'%H:%M:%S  %m-%d-%Y',

            6  => 'j F Y, H:i:s',       // '%e %B %Y, %H:%M:%S',
            7  => 'd.m.Y - H:i:s',      // '%d.%m.%Y - %H:%M:%S',
            9  => 'd.m.Y - h:i:s A',    // '%d.%m.%Y - %l:%M:%S %p',
            10 => 'H:i:s - d.m.Y',      // '%H:%M:%S - %d.%m.%Y', //Default date format
            11 => 'H:i - d.m.Y',        // '%H:%M - %d.%m.%Y',

            12 => 'Y-m-d H:i',          // '%Y-%m-%d %H:%M',
            13 => 'Y-m-d H:i:s',        // '%Y-%m-%d %H:%M:%S'
        ];
    }


    /**
     * @return false|string
     */
    public function uploadProfilePicture() {
        $path = WWW_ROOT . 'img' . DS . 'userimages' . DS;

        if ($_FILES['Picture']['error'] !== UPLOAD_ERR_OK) {
            return false;
        }

        if (!is_dir($path)) {
            return false;
        }

        $tmpImage = UUID::v4();
        if (move_uploaded_file($_FILES['Picture']['tmp_name'], $path . $tmpImage)) {
            $newImage = UUID::v4() . '.png';

            //Try to create new user image from uploaded image

            $tmpImageFull = $path . $tmpImage;
            $newImageFull = $path . $newImage;

            $imgsize = getimagesize($tmpImageFull);
            $width = $imgsize[0];
            $height = $imgsize[1];
            $imgtype = $imgsize[2];

            switch ($imgtype) {
                /**
                 * 1 => GIF
                 * 2 => JPG
                 * 3 => PNG
                 * 4 => SWF
                 * 5 => PSD
                 * 6 => BMP
                 * 7 => TIFF(intel byte order)
                 * 8 => TIFF(motorola byte order)
                 * 9 => JPC
                 * 10 => JP2
                 * 11 => JPX
                 * 12 => JB2
                 * 13 => SWC
                 * 14 => IFF
                 * 15 => WBMP
                 * 16 => XBM
                 */
                case 1:
                    $srcImg = imagecreatefromgif($tmpImageFull);
                    break;
                case 2:
                    $srcImg = imagecreatefromjpeg($tmpImageFull);
                    break;
                case 3:
                    $srcImg = imagecreatefrompng($tmpImageFull);
                    break;
                default:
                    //Filetype not supported!
                    return false;
                    break;
            }

            $newWidth = 240;
            $newHeight = 240;
            //Thanks to http://php.net/manual/de/function.imagecopyresized.php#50019 :)
            if ($width > $height && $newWidth < $newHeight) {
                $newHeight = $height / ($width / $newWidth);
            } else if ($width < $height && $newHeight < $width) {
                $newWidth = $width / ($height / $newHeight);
            } else {
                $newHeight = $height;
                $newWidth = $width;
            }
            $destImg = imagecreatetruecolor($newWidth, $newHeight);
            $transparent = imagecolorallocatealpha($destImg, 0, 0, 0, 127);
            imagefill($destImg, 0, 0, $transparent);
            imagecopyresized($destImg, $srcImg, 0, 0, 0, 0, $newWidth, $newHeight, $width, $height);
            imagealphablending($destImg, false);
            imagesavealpha($destImg, true);
            imagepng($destImg, $newImageFull);
            imagedestroy($destImg);

            //Delete source image
            unlink($tmpImageFull);
            return $newImage;
        }
        return false;
    }

}
